package com.sun.xml.internal.ws.client.dispatch;

import com.sun.xml.internal.ws.client.ResponseContext;
import java.util.concurrent.Callable;
import com.sun.xml.internal.ws.api.pipe.Fiber;
import javax.xml.transform.Source;
import java.util.Iterator;
import com.sun.xml.internal.ws.api.message.Attachment;
import java.util.List;
import com.sun.xml.internal.ws.message.DataHandlerAttachment;
import javax.activation.DataHandler;
import java.util.Map;
import java.util.ArrayList;
import com.sun.xml.internal.ws.message.AttachmentSetImpl;
import java.util.HashMap;
import com.sun.xml.internal.ws.api.message.AttachmentSet;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URISyntaxException;
import java.net.URI;
import com.sun.xml.internal.ws.api.BindingID;
import com.sun.istack.internal.NotNull;
import com.sun.xml.internal.ws.api.WSBinding;
import com.sun.xml.internal.ws.api.message.Message;
import javax.xml.ws.WebServiceException;
import javax.xml.bind.JAXBException;
import com.sun.xml.internal.ws.encoding.soap.DeserializationException;
import com.sun.xml.internal.ws.resources.DispatchMessages;
import javax.xml.ws.soap.SOAPFaultException;
import com.sun.xml.internal.ws.fault.SOAPFaultBuilder;
import com.sun.xml.internal.ws.client.ResponseContextReceiver;
import com.sun.xml.internal.ws.client.RequestContext;
import java.util.concurrent.Future;
import javax.xml.ws.AsyncHandler;
import com.sun.xml.internal.ws.api.addressing.AddressingVersion;
import com.sun.xml.internal.ws.api.message.AddressingUtils;
import com.sun.xml.internal.ws.client.AsyncInvoker;
import com.sun.xml.internal.ws.api.server.Container;
import com.sun.xml.internal.ws.client.AsyncResponseImpl;
import java.util.logging.Level;
import com.sun.xml.internal.ws.api.server.ContainerResolver;
import javax.xml.ws.Response;
import com.sun.xml.internal.ws.api.message.Packet;
import com.sun.xml.internal.ws.api.client.WSPortInfo;
import com.sun.istack.internal.Nullable;
import com.sun.xml.internal.ws.api.addressing.WSEndpointReference;
import com.sun.xml.internal.ws.binding.BindingImpl;
import com.sun.xml.internal.ws.api.pipe.Tube;
import com.sun.xml.internal.ws.client.WSServiceDelegate;
import javax.xml.namespace.QName;
import com.sun.xml.internal.ws.api.SOAPVersion;
import javax.xml.ws.Service;
import java.util.logging.Logger;
import javax.xml.ws.Dispatch;
import com.sun.xml.internal.ws.client.Stub;

public abstract class DispatchImpl<T> extends Stub implements Dispatch<T>
{
    private static final Logger LOGGER;
    final Service.Mode mode;
    final SOAPVersion soapVersion;
    final boolean allowFaultResponseMsg;
    static final long AWAIT_TERMINATION_TIME = 800L;
    static final String HTTP_REQUEST_METHOD_GET = "GET";
    static final String HTTP_REQUEST_METHOD_POST = "POST";
    static final String HTTP_REQUEST_METHOD_PUT = "PUT";
    
    @Deprecated
    protected DispatchImpl(QName port, Service.Mode mode, WSServiceDelegate owner, Tube pipe, BindingImpl binding, @Nullable WSEndpointReference epr) {
        super(port, owner, pipe, binding, (owner.getWsdlService() != null) ? owner.getWsdlService().get(port) : null, owner.getEndpointAddress(port), epr);
        this.mode = mode;
        this.soapVersion = binding.getSOAPVersion();
        this.allowFaultResponseMsg = false;
    }
    
    protected DispatchImpl(WSPortInfo portInfo, Service.Mode mode, BindingImpl binding, @Nullable WSEndpointReference epr) {
        this(portInfo, mode, binding, epr, false);
    }
    
    protected DispatchImpl(WSPortInfo portInfo, Service.Mode mode, BindingImpl binding, @Nullable WSEndpointReference epr, boolean allowFaultResponseMsg) {
        this(portInfo, mode, binding, null, epr, allowFaultResponseMsg);
    }
    
    protected DispatchImpl(WSPortInfo portInfo, Service.Mode mode, BindingImpl binding, Tube pipe, @Nullable WSEndpointReference epr, boolean allowFaultResponseMsg) {
        super(portInfo, binding, pipe, portInfo.getEndpointAddress(), epr);
        this.mode = mode;
        this.soapVersion = binding.getSOAPVersion();
        this.allowFaultResponseMsg = allowFaultResponseMsg;
    }
    
    protected DispatchImpl(WSPortInfo portInfo, Service.Mode mode, Tube pipe, BindingImpl binding, @Nullable WSEndpointReference epr, boolean allowFaultResponseMsg) {
        super(portInfo, binding, pipe, portInfo.getEndpointAddress(), epr);
        this.mode = mode;
        this.soapVersion = binding.getSOAPVersion();
        this.allowFaultResponseMsg = allowFaultResponseMsg;
    }
    
    abstract Packet createPacket(T p0);
    
    abstract T toReturnValue(Packet p0);
    
    public final Response<T> invokeAsync(T param) {
        Container old = ContainerResolver.getDefault().enterContainer(this.owner.getContainer());
        try {
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                this.dumpParam(param, "invokeAsync(T)");
            }
            AsyncInvoker invoker = new DispatchImpl.DispatchAsyncInvoker(param);
            AsyncResponseImpl<T> ft = new AsyncResponseImpl(invoker, null);
            invoker.setReceiver(ft);
            ft.run();
            return ft;
        }
        finally {
            ContainerResolver.getDefault().exitContainer(old);
        }
    }
    
    private void dumpParam(T param, String method) {
        if ((param instanceof Packet)) {
            Packet message = (Packet)param;
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                AddressingVersion av = this.getBinding().getAddressingVersion();
                SOAPVersion sv = this.getBinding().getSOAPVersion();
                String action = (av != null && message.getMessage() != null) ? AddressingUtils.getAction(message.getMessage().getHeaders(), av, sv) : null;
                String msgId = (av != null && message.getMessage() != null) ? AddressingUtils.getMessageID(message.getMessage().getHeaders(), av, sv) : null;
                DispatchImpl.LOGGER.fine("In DispatchImpl." + method + " for message with action: " + action + " and msg ID: " + msgId + " msg: " + message.getMessage());
                if (message.getMessage() == null) {
                    DispatchImpl.LOGGER.fine("Dispatching null message for action: " + action + " and msg ID: " + msgId);
                }
            }
        }
    }
    
    public final Future<?> invokeAsync(T param, AsyncHandler<T> asyncHandler) {
        Container old = ContainerResolver.getDefault().enterContainer(this.owner.getContainer());
        try {
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                this.dumpParam(param, "invokeAsync(T, AsyncHandler<T>)");
            }
            AsyncInvoker invoker = new DispatchImpl.DispatchAsyncInvoker(param);
            AsyncResponseImpl<T> ft = new AsyncResponseImpl(invoker, asyncHandler);
            invoker.setReceiver(ft);
            invoker.setNonNullAsyncHandlerGiven(asyncHandler != null);
            ft.run();
            return ft;
        }
        finally {
            ContainerResolver.getDefault().exitContainer(old);
        }
    }
    
    public final T doInvoke(T in, RequestContext rc, ResponseContextReceiver receiver) {
        Packet response = null;
        try {
            try {
                checkNullAllowed(in, rc, this.binding, this.mode);
                Packet message = this.createPacket(in);
                message.setState(Packet.State.ClientRequest);
                this.resolveEndpointAddress(message, rc);
                this.setProperties(message, true);
                response = this.process(message, rc, receiver);
                Message msg = response.getMessage();
                if (msg != null && msg.isFault() && !this.allowFaultResponseMsg) {
                    SOAPFaultBuilder faultBuilder = SOAPFaultBuilder.create(msg);
                    throw (SOAPFaultException)faultBuilder.createException(null);
                }
            }
            catch (JAXBException e) {
                throw new DeserializationException(DispatchMessages.INVALID_RESPONSE_DESERIALIZATION(), new Object[] { e });
            }
            catch (WebServiceException e2) {
                throw e2;
            }
            catch (Throwable e3) {
                throw new WebServiceException(e3);
            }
            Object returnValue = this.toReturnValue(response);
        }
        finally {
            if (response != null && response.transportBackChannel != null) {
                response.transportBackChannel.close();
            }
        }
    }
    
    public final T invoke(T in) {
        Container old = ContainerResolver.getDefault().enterContainer(this.owner.getContainer());
        try {
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                this.dumpParam(in, "invoke(T)");
            }
            return this.doInvoke(in, this.requestContext, this);
        }
        finally {
            ContainerResolver.getDefault().exitContainer(old);
        }
    }
    
    public final void invokeOneWay(T in) {
        Container old = ContainerResolver.getDefault().enterContainer(this.owner.getContainer());
        try {
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                this.dumpParam(in, "invokeOneWay(T)");
            }
            try {
                checkNullAllowed(in, this.requestContext, this.binding, this.mode);
                Packet request = this.createPacket(in);
                request.setState(Packet.State.ClientRequest);
                this.setProperties(request, false);
                this.process(request, this.requestContext, this);
            }
            catch (WebServiceException e) {
                throw e;
            }
            catch (Throwable e2) {
                throw new WebServiceException(e2);
            }
        }
        finally {
            ContainerResolver.getDefault().exitContainer(old);
        }
    }
    
    void setProperties(Packet packet, boolean expectReply) {
        packet.expectReply = Boolean.valueOf(expectReply);
    }
    
    static boolean isXMLHttp(@NotNull WSBinding binding) {
        return binding.getBindingId().equals(BindingID.XML_HTTP);
    }
    
    static boolean isPAYLOADMode(@NotNull Service.Mode mode) {
        return mode == Service.Mode.PAYLOAD;
    }
    
    static void checkNullAllowed(@Nullable Object in, RequestContext rc, WSBinding binding, Service.Mode mode) {
        if (in != null) {
            return;
        }
        if (isXMLHttp(binding)) {
            if (methodNotOk(rc)) {
                throw new WebServiceException(DispatchMessages.INVALID_NULLARG_XMLHTTP_REQUEST_METHOD("POST", "GET"));
            }
        }
        else if (mode == Service.Mode.MESSAGE) {
            throw new WebServiceException(DispatchMessages.INVALID_NULLARG_SOAP_MSGMODE(mode.name(), Service.Mode.PAYLOAD.toString()));
        }
        return;
    }
    
    static boolean methodNotOk(@NotNull RequestContext rc) {
        String requestMethod = (String)rc.get("javax.xml.ws.http.request.method");
        String request = (requestMethod == null) ? "POST" : requestMethod;
        return "POST".equalsIgnoreCase(request) || "PUT".equalsIgnoreCase(request);
    }
    
    public static void checkValidSOAPMessageDispatch(WSBinding binding, Service.Mode mode) {
        if (isXMLHttp(binding)) {
            throw new WebServiceException(DispatchMessages.INVALID_SOAPMESSAGE_DISPATCH_BINDING("http://www.w3.org/2004/08/wsdl/http", "http://schemas.xmlsoap.org/wsdl/soap/http or http://www.w3.org/2003/05/soap/bindings/HTTP/"));
        }
        if (isPAYLOADMode(mode)) {
            throw new WebServiceException(DispatchMessages.INVALID_SOAPMESSAGE_DISPATCH_MSGMODE(mode.name(), Service.Mode.MESSAGE.toString()));
        }
        return;
    }
    
    public static void checkValidDataSourceDispatch(WSBinding binding, Service.Mode mode) {
        if (!isXMLHttp(binding)) {
            throw new WebServiceException(DispatchMessages.INVALID_DATASOURCE_DISPATCH_BINDING("SOAP/HTTP", "http://www.w3.org/2004/08/wsdl/http"));
        }
        if (isPAYLOADMode(mode)) {
            throw new WebServiceException(DispatchMessages.INVALID_DATASOURCE_DISPATCH_MSGMODE(mode.name(), Service.Mode.MESSAGE.toString()));
        }
        return;
    }
    
    @NotNull
    public final QName getPortName() {
        return this.portname;
    }
    
    void resolveEndpointAddress(@NotNull Packet message, @NotNull RequestContext requestContext) {
        boolean p = message.packetTakesPriorityOverRequestContext;
        String endpoint;
        if (p && message.endpointAddress != null) {
            endpoint = message.endpointAddress.toString();
        }
        else {
            endpoint = (String)requestContext.get("javax.xml.ws.service.endpoint.address");
        }
        if (endpoint == null) {
            if (message.endpointAddress == null) {
                throw new WebServiceException(DispatchMessages.INVALID_NULLARG_URI());
            }
            endpoint = message.endpointAddress.toString();
        }
        String pathInfo = null;
        String queryString = null;
        if (p && message.invocationProperties.get("javax.xml.ws.http.request.pathinfo") != null) {
            pathInfo = (String)message.invocationProperties.get("javax.xml.ws.http.request.pathinfo");
        }
        else if (requestContext.get("javax.xml.ws.http.request.pathinfo") != null) {
            pathInfo = (String)requestContext.get("javax.xml.ws.http.request.pathinfo");
        }
        if (p && message.invocationProperties.get("javax.xml.ws.http.request.querystring") != null) {
            queryString = (String)message.invocationProperties.get("javax.xml.ws.http.request.querystring");
        }
        else if (requestContext.get("javax.xml.ws.http.request.querystring") != null) {
            queryString = (String)requestContext.get("javax.xml.ws.http.request.querystring");
        }
        if (pathInfo != null || queryString != null) {
            pathInfo = checkPath(pathInfo);
            queryString = checkQuery(queryString);
            if (endpoint != null) {
                try {
                    URI endpointURI = new URI(endpoint);
                    endpoint = this.resolveURI(endpointURI, pathInfo, queryString);
                }
                catch (URISyntaxException e) {
                    throw new WebServiceException(DispatchMessages.INVALID_URI(endpoint));
                }
            }
        }
        requestContext.put("javax.xml.ws.service.endpoint.address", endpoint);
    }
    
    @NotNull
    protected String resolveURI(@NotNull URI endpointURI, @Nullable String pathInfo, @Nullable String queryString) {
        String query = null;
        String fragment = null;
        if (queryString != null) {
            URI result;
            try {
                URI tp = new URI(null, null, endpointURI.getPath(), queryString, null);
                result = endpointURI.resolve(tp);
            }
            catch (URISyntaxException e) {
                throw new WebServiceException(DispatchMessages.INVALID_QUERY_STRING(queryString));
            }
            query = result.getQuery();
            fragment = result.getFragment();
        }
        String path = (pathInfo != null) ? pathInfo : endpointURI.getPath();
        try {
            StringBuilder spec = new StringBuilder();
            if (path != null) {
                spec.append(path);
            }
            if (query != null) {
                spec.append("?");
                spec.append(query);
            }
            if (fragment != null) {
                spec.append("#");
                spec.append(fragment);
            }
            return new URL(endpointURI.toURL(), spec.toString()).toExternalForm();
        }
        catch (MalformedURLException e2) {
            throw new WebServiceException(DispatchMessages.INVALID_URI_RESOLUTION(path));
        }
    }
    
    private static String checkPath(@Nullable String path) {
        return (path == null || path.startsWith("/")) ? path : ("/" + path);
    }
    
    private static String checkQuery(@Nullable String query) {
        if (query == null) {
            return null;
        }
        if (query.indexOf(63) == 0) {
            throw new WebServiceException(DispatchMessages.INVALID_QUERY_LEADING_CHAR(query));
        }
        return query;
    }
    
    protected AttachmentSet setOutboundAttachments() {
        HashMap<String, DataHandler> attachments = (HashMap)this.getRequestContext().get("javax.xml.ws.binding.attachments.outbound");
        if (attachments != null) {
            List<Attachment> alist = new ArrayList();
            for (Map.Entry<String, DataHandler> att : attachments.entrySet()) {
                DataHandlerAttachment dha = new DataHandlerAttachment((String)att.getKey(), (DataHandler)att.getValue());
                alist.add(dha);
            }
            return new AttachmentSetImpl(alist);
        }
        return new AttachmentSetImpl();
    }
    
    public void setOutboundHeaders(Object... headers) {
        throw new UnsupportedOperationException();
    }
    
    @Deprecated
    public static Dispatch<Source> createSourceDispatch(QName port, Service.Mode mode, WSServiceDelegate owner, Tube pipe, BindingImpl binding, WSEndpointReference epr) {
        if (isXMLHttp(binding)) {
            return new RESTSourceDispatch(port, mode, owner, pipe, binding, epr);
        }
        return new SOAPSourceDispatch(port, mode, owner, pipe, binding, epr);
    }
    
    public static Dispatch<Source> createSourceDispatch(WSPortInfo portInfo, Service.Mode mode, BindingImpl binding, WSEndpointReference epr) {
        if (isXMLHttp(binding)) {
            return new RESTSourceDispatch(portInfo, mode, binding, epr);
        }
        return new SOAPSourceDispatch(portInfo, mode, binding, epr);
    }
    
    static void access$100(DispatchImpl x0, Object x1, String x2) {
        x0.dumpParam(x1, x2);
    }
    
    static void access$1000(DispatchImpl x0, AsyncResponseImpl x1, Packet x2, RequestContext x3, Fiber.CompletionCallback x4) {
        x0.processAsync(x1, x2, x3, x4);
    }
    
    static {
        LOGGER = Logger.getLogger(DispatchImpl.class.getName());
    }
    
    private class Invoker implements Callable
    {
        private final T param;
        private final RequestContext rc;
        private ResponseContextReceiver receiver;
        
        Invoker(T param) {
            this.rc = DispatchImpl.this.requestContext.copy();
            this.param = param;
        }
        
        @Override
        public T call() throws Exception {
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                DispatchImpl.access$100(DispatchImpl.this, this.param, "call()");
            }
            return DispatchImpl.this.doInvoke(this.param, this.rc, this.receiver);
        }
        
        void setReceiver(ResponseContextReceiver receiver) {
            this.receiver = receiver;
        }
    }
    
    private class DispatchAsyncInvoker extends AsyncInvoker
    {
        private final T param;
        private final RequestContext rc;
        
        DispatchAsyncInvoker(T param) {
            this.rc = DispatchImpl.this.requestContext.copy();
            this.param = param;
        }
        
        public void do_run() {
            DispatchImpl.checkNullAllowed(this.param, this.rc, DispatchImpl.this.binding, DispatchImpl.this.mode);
            Packet message = DispatchImpl.this.createPacket(this.param);
            message.setState(Packet.State.ClientRequest);
            message.nonNullAsyncHandlerGiven = Boolean.valueOf(this.nonNullAsyncHandlerGiven);
            DispatchImpl.this.resolveEndpointAddress(message, this.rc);
            DispatchImpl.this.setProperties(message, true);
            String action = null;
            String msgId = null;
            if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                AddressingVersion av = DispatchImpl.this.getBinding().getAddressingVersion();
                SOAPVersion sv = DispatchImpl.this.getBinding().getSOAPVersion();
                action = ((av != null && message.getMessage() != null) ? AddressingUtils.getAction(message.getMessage().getHeaders(), av, sv) : null);
                msgId = ((av != null && message.getMessage() != null) ? AddressingUtils.getMessageID(message.getMessage().getHeaders(), av, sv) : null);
                DispatchImpl.LOGGER.fine("In DispatchAsyncInvoker.do_run for async message with action: " + action + " and msg ID: " + msgId);
            }
            String actionUse = action;
            String msgIdUse = msgId;
            Fiber.CompletionCallback callback = new Fiber.CompletionCallback() {
                public void onCompletion(@NotNull Packet response) {
                    if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                        DispatchImpl.LOGGER.fine("Done with processAsync in DispatchAsyncInvoker.do_run, and setting response for async message with action: " + actionUse + " and msg ID: " + msgIdUse);
                    }
                    Message msg = response.getMessage();
                    if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                        DispatchImpl.LOGGER.fine("Done with processAsync in DispatchAsyncInvoker.do_run, and setting response for async message with action: " + actionUse + " and msg ID: " + msgIdUse + " msg: " + msg);
                    }
                    try {
                        if (msg != null && msg.isFault() && !DispatchImpl.this.allowFaultResponseMsg) {
                            SOAPFaultBuilder faultBuilder = SOAPFaultBuilder.create(msg);
                            throw (SOAPFaultException)faultBuilder.createException(null);
                        }
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.setResponseContext(new ResponseContext(response));
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(DispatchImpl.this.toReturnValue(response), null);
                    }
                    catch (JAXBException e) {
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(null, new DeserializationException(DispatchMessages.INVALID_RESPONSE_DESERIALIZATION(), new Object[] { e }));
                    }
                    catch (WebServiceException e2) {
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(null, e2);
                    }
                    catch (Throwable e3) {
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(null, new WebServiceException(e3));
                    }
                }
                
                public void onCompletion(@NotNull Throwable error) {
                    if (DispatchImpl.LOGGER.isLoggable(Level.FINE)) {
                        DispatchImpl.LOGGER.fine("Done with processAsync in DispatchAsyncInvoker.do_run, and setting response for async message with action: " + actionUse + " and msg ID: " + msgIdUse + " Throwable: " + error.toString());
                    }
                    if ((error instanceof WebServiceException)) {
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(null, error);
                    }
                    else {
                        DispatchImpl.DispatchAsyncInvoker.this.responseImpl.set(null, new WebServiceException(error));
                    }
                }
            };
            DispatchImpl.access$1000(DispatchImpl.this, this.responseImpl, message, this.rc, callback);
        }
    }
}
